package de.fhdw.wtf.common.ast;

import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;

import de.fhdw.wtf.common.ast.visitor.GroupElementExceptionVisitor;
import de.fhdw.wtf.common.ast.visitor.GroupElementReturnVisitor;
import de.fhdw.wtf.common.ast.visitor.GroupElementVisitor;
import de.fhdw.wtf.common.constants.parser.AstDescriptionConstants;
import de.fhdw.wtf.common.token.Token;

/**
 * This class represents a {@link Group} in an abstract syntax tree of the <code>WTF-Language</code>. A {@link Group} is
 * either contained by a {@link Model} or by another {@link Group} and can contain multiple {@link GroupElement}s by
 * itself.
 */
public final class Group extends SyntaxObject implements GroupElement {
	
	/**
	 * generated.
	 */
	private static final long serialVersionUID = 1603342875497690969L;
	
	/**
	 * Constant for Line-Break.
	 */
	private static final String BREAK = " \n";
	
	/**
	 * The Elements of the group.
	 */
	private final Collection<GroupElement> groupElements;
	
	/**
	 * The name of the group.
	 */
	private final Name name;
	
	/**
	 * Private-Constructor of {@link Group}.
	 * 
	 * @param name
	 *            name
	 * @param groups
	 *            Elements of the group
	 * @param firstToken
	 *            firstToken
	 * @param lastToken
	 *            lastToken
	 */
	private Group(final Name name, final Collection<GroupElement> groups, final Token firstToken, final Token lastToken) {
		super(firstToken, lastToken);
		this.groupElements = groups;
		this.name = name;
	}
	
	/**
	 * Creates a {@link Group}-Object.
	 * 
	 * @param name
	 *            name
	 * @param groups
	 *            Elements of the group
	 * @param firstToken
	 *            firstToken
	 * @param lastToken
	 *            lastToken
	 * @return The Group-Object.
	 */
	public static Group create(final Name name,
			final Collection<GroupElement> groups,
			final Token firstToken,
			final Token lastToken) {
		return new Group(name, groups, firstToken, lastToken);
	}
	
	/**
	 * Creates a {@link Group}-Object.
	 * 
	 * @param name
	 *            name
	 * @param groups
	 *            Elements of the group
	 * @param firstToken
	 *            firstToken
	 * @return The Group-Object.
	 */
	public static Group create(final Name name, final Collection<GroupElement> groups, final Token firstToken) {
		return new Group(name, groups, firstToken, null);
	}
	
	@Override
	public boolean equals(final Object o) {
		if (o instanceof Group) {
			boolean result = true;
			final Group other = (Group) o;
			result =
					result && this.getName().equals(other.getName())
							&& this.getGroupElements().size() == other.getGroupElements().size();
			if (result) {
				final GroupElement[] thisArray =
						this.getGroupElements().toArray(new GroupElement[this.getGroupElements().size()]);
				final GroupElement[] otherArray =
						other.getGroupElements().toArray(new GroupElement[other.getGroupElements().size()]);
				for (int i = 0; i < this.getGroupElements().size(); i++) {
					result = result && thisArray[i].equals(otherArray[i]);
				}
				return result;
			}
		}
		return false;
	}
	
	@Override
	public int hashCode() {
		int result = this.getName().hashCode();
		for (final GroupElement element : this.getGroupElements()) {
			result ^= element.hashCode();
		}
		return result;
	}
	
	@Override
	public String toString() {
		final StringBuilder result = new StringBuilder();
		result.append(this.getName().toString()).append(AstDescriptionConstants.COLON_TOKEN)
				.append(AstDescriptionConstants.GROUP_TOKEN).append(AstDescriptionConstants.EQUAL_TOKEN)
				.append(AstDescriptionConstants.SQUARE_BRACKET_OPEN).append(BREAK);
		final Iterator<GroupElement> i = this.getGroupElements().iterator();
		while (i.hasNext()) {
			final GroupElement current = i.next();
			result.append(current.toString()).append(BREAK);
		}
		result.append(AstDescriptionConstants.SQUARE_BRACKET_CLOSE).append(AstDescriptionConstants.SEMICOLON_TOKEN);
		return result.toString();
	}
	
	@Override
	public void accept(final GroupElementVisitor visitor) {
		visitor.handle(this);
	}
	
	@Override
	public <X extends Exception> void accept(final GroupElementExceptionVisitor<X> visitor) throws X {
		visitor.handle(this);
	}
	
	@Override
	public <X> X accept(final GroupElementReturnVisitor<X> visitor) {
		return visitor.handle(this);
	}
	
	// ///// PROJECTIONS ///////////
	
	/**
	 * Returns a collection of {@link GroupElement}. The collection can't be modified.
	 * 
	 * @return the groups
	 */
	public Collection<GroupElement> getGroupElements() {
		return Collections.unmodifiableCollection(this.groupElements);
	}
	
	@Override
	public Name getName() {
		return this.name;
	}
	
}
