package de.fhdw.wtf.walker.tasks;

import java.util.HashSet;
import java.util.Set;

import de.fhdw.wtf.common.ast.Model;
import de.fhdw.wtf.common.ast.type.AtomicType;
import de.fhdw.wtf.common.ast.type.CompositeType;
import de.fhdw.wtf.common.ast.type.ListType;
import de.fhdw.wtf.common.ast.type.MapType;
import de.fhdw.wtf.common.ast.type.ProductElementType;
import de.fhdw.wtf.common.ast.type.ProductType;
import de.fhdw.wtf.common.ast.type.SumType;
import de.fhdw.wtf.common.ast.type.ThrownType;
import de.fhdw.wtf.common.ast.type.Type;
import de.fhdw.wtf.common.ast.type.TypeProxy;
import de.fhdw.wtf.common.ast.visitor.CompositeTypeVisitorException;
import de.fhdw.wtf.common.ast.visitor.TypeVisitorException;
import de.fhdw.wtf.common.exception.walker.InvalidProductException;
import de.fhdw.wtf.common.exception.walker.TaskException;
import de.fhdw.wtf.common.task.TaskExecutor;
import de.fhdw.wtf.walker.walker.SimpleWalkerTaskForTypes;

/**
 * This Task check that the names of the productElements are pairwise distinct in every product!
 */
public final class ProductCheck extends SimpleWalkerTaskForTypes {
	
	/**
	 * Constructor of {@link ProductCheck}.
	 * 
	 * @param m
	 *            model
	 * @param taskmanager
	 *            taskmanager
	 */
	private ProductCheck(final Model m, final TaskExecutor taskmanager) {
		super(m, taskmanager);
	}
	
	/**
	 * Creates a {@link ProductCheck}-Object.
	 * 
	 * @param m
	 *            model
	 * @param taskmanager
	 *            taskmanager
	 * @return The {@link ProductCheck}-Object.
	 */
	public static ProductCheck create(final Model m, final TaskExecutor taskmanager) {
		return new ProductCheck(m, taskmanager);
	}
	
	@Override
	public void handleType(final Type type) throws TaskException {
		type.accept(new TypeVisitorException<TaskException>() {
			
			@Override
			public void handle(final AtomicType s) throws TaskException {
				// Nothing to do
			}
			
			@Override
			public void handle(final CompositeType c) throws TaskException {
				c.accept(new CompositeTypeVisitorException<TaskException>() {
					
					@Override
					public void handle(final MapType map) throws TaskException {
						// Nothing to do
					}
					
					@Override
					public void handle(final ListType list) throws TaskException {
						// Nothing to do
					}
					
					@Override
					public void handle(final ProductType product) throws TaskException {
						final Set<String> set = new HashSet<>();
						for (final ProductElementType element : product.getElements()) {
							if (set.contains(element.getName())) {
								throw InvalidProductException.create(
										element.getName(),
										product.getFirstToken(),
										product.getLastToken());
							}
							set.add(element.getName());
						}
					}
					
					@Override
					public void handle(final SumType sum) throws TaskException {
						// Nothing to do
					}
					
					@Override
					public void handle(final ThrownType thrownType) throws TaskException {
						// Nothing to do
					}
				});
			}
			
			@Override
			public void handle(final TypeProxy s) throws TaskException {
				// Nothing to do
			}
		});
	}
	
	@Override
	public void finalizeTask() throws TaskException {
		// Nothing to do here
	}
	
	@Override
	public void beginTask() throws TaskException {
		// Nothing to do here
	}
}
