package de.fhdw.wtf.walker.tasks;

import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import de.fhdw.wtf.common.ast.Attribute;
import de.fhdw.wtf.common.ast.Constructor;
import de.fhdw.wtf.common.ast.ConstructorByReferenceState;
import de.fhdw.wtf.common.ast.ConstructorByTypeAndSignatureState;
import de.fhdw.wtf.common.ast.ConstructorInvalidState;
import de.fhdw.wtf.common.ast.ConstructorOrOperation;
import de.fhdw.wtf.common.ast.ConstructorReference;
import de.fhdw.wtf.common.ast.ConstructorReferenceState;
import de.fhdw.wtf.common.ast.Group;
import de.fhdw.wtf.common.ast.Model;
import de.fhdw.wtf.common.ast.Operation;
import de.fhdw.wtf.common.ast.type.AtomicType;
import de.fhdw.wtf.common.ast.type.ClassType;
import de.fhdw.wtf.common.ast.type.CompositeType;
import de.fhdw.wtf.common.ast.type.ListType;
import de.fhdw.wtf.common.ast.type.MapType;
import de.fhdw.wtf.common.ast.type.ProductType;
import de.fhdw.wtf.common.ast.type.SumType;
import de.fhdw.wtf.common.ast.type.ThrownType;
import de.fhdw.wtf.common.ast.type.Type;
import de.fhdw.wtf.common.ast.type.TypeProxy;
import de.fhdw.wtf.common.ast.visitor.CompositeTypeVisitorReturn;
import de.fhdw.wtf.common.ast.visitor.ConstructorOrOperationExceptionVisitor;
import de.fhdw.wtf.common.ast.visitor.ConstructorReferenceStateVisitorReturnException;
import de.fhdw.wtf.common.ast.visitor.TypeVisitorReturn;
import de.fhdw.wtf.common.exception.walker.InvalidConstructorReferenceListException;
import de.fhdw.wtf.common.exception.walker.TaskException;
import de.fhdw.wtf.common.task.TaskExecutor;
import de.fhdw.wtf.walker.walker.SimpleWalkerTask;

/**
 * Checks if the list of super constructors in constructors apply to the rules for constructors. Should not run before
 * {@link ConstructorReferencer}.
 */
public final class ReferencedConstructorsCheck extends SimpleWalkerTask {
	
	/**
	 * The instantiations classes must do.
	 */
	private final Map<Type, List<Type>> necessaryCalls;
	
	/**
	 * Returns a new instance of {@link ReferencedConstructorsCheck}.
	 * 
	 * @param model
	 *            The model to check.
	 * @param taskmanager
	 *            The executor which performs this task.
	 * @return a new instance of {@link ReferencedConstructorsCheck}
	 */
	public static ReferencedConstructorsCheck create(final Model model, final TaskExecutor taskmanager) {
		return new ReferencedConstructorsCheck(model, taskmanager);
	}
	
	/**
	 * Instantiates a new instance of {@link ReferencedConstructorsCheck}.
	 * 
	 * @param model
	 *            The model to check.
	 * @param taskmanager
	 *            The executor which performs this task.
	 */
	private ReferencedConstructorsCheck(final Model model, final TaskExecutor taskmanager) {
		super(model, taskmanager);
		this.necessaryCalls = model.getConstructorCallDependencies();
	}
	
	@Override
	public void handleClass(final ClassType c) throws TaskException {
		// Nothing to do.
	}
	
	@Override
	public void handleGroup(final Group g) throws TaskException {
		// Nothing to do.
	}
	
	@Override
	public void handleAttribute(final Attribute a, final ClassType owner) throws TaskException {
		// Nothing to do.
	}
	
	@Override
	public void handleConstructorOrOperation(final ConstructorOrOperation coo, final ClassType owner)
			throws TaskException {
		coo.accept(new ConstructorOrOperationExceptionVisitor<TaskException>() {
			
			@Override
			public void handleConstructor(final Constructor constructor) throws TaskException {
				final Type caller = constructor.getReturnType();
				final List<Type> typesToCall = ReferencedConstructorsCheck.this.necessaryCalls.get(caller);
				final Collection<Type> calledSuperTypes = new LinkedList<>();
				final Iterator<ConstructorReference> references = constructor.getSuperConstructors().iterator();
				while (references.hasNext()) {
					final ConstructorReference reference = references.next();
					calledSuperTypes.add(ReferencedConstructorsCheck.this.getReferencedType(reference.getState()));
				}
				if (!ReferencedConstructorsCheck.this.checkIfLegitimeSuperTypeList(
						caller,
						calledSuperTypes,
						typesToCall)) {
					throw InvalidConstructorReferenceListException.create(
							constructor,
							"Either not all constructors that have to be called are being referenced"
									+ "or too many constructors are being referenced!"
									+ "\nSupertypes which have to be instanciated by this class:"
									+ typesToCall.toString() + "\nCalling constructor: ");
				}
				
			}
			
			@Override
			public void handleOperation(final Operation operation) throws TaskException {
				// Nothing to do.
			}
			
		});
		
	}
	
	/**
	 * Checks if the two collections contain each over mutually.
	 * 
	 * @param caller
	 *            the caller.
	 * @param calledSuperTypes
	 *            one collection.
	 * @param typesToCall
	 *            one collection.
	 * @return true if they contain each other. Otherwise false.
	 */
	protected boolean checkIfLegitimeSuperTypeList(final Type caller,
			final Collection<Type> calledSuperTypes,
			final Collection<Type> typesToCall) {
		return this.checkOneWay(caller, calledSuperTypes, typesToCall)
				&& this.checkOneWay(caller, typesToCall, calledSuperTypes);
	}
	
	/**
	 * Checks if all elements in left are contained in right.
	 * 
	 * @param caller
	 *            the caller.
	 * 
	 * @param left
	 *            a collection.
	 * @param right
	 *            a collection.
	 * @return true if right contains all elements of right.
	 */
	private boolean checkOneWay(final Type caller, final Collection<Type> left, final Collection<Type> right) {
		final Iterator<Type> leftIterator = left.iterator();
		boolean legitimate = true;
		while (legitimate && leftIterator.hasNext()) {
			final Type currentCalled = leftIterator.next();
			legitimate = currentCalled.accept(new TypeChecker(caller, currentCalled, right));
		}
		return legitimate;
	}
	
	/**
	 * Checks if a type is part of a given list of types.
	 */
	class TypeChecker implements TypeVisitorReturn<Boolean> {
		
		/**
		 * Calling type.
		 */
		private final Type caller;
		
		/**
		 * Called type.
		 */
		private final Type called;
		/**
		 * types to call.
		 */
		private final Collection<Type> typesToCall;
		
		/**
		 * Returns a new TypeChecker.
		 * 
		 * @param caller
		 *            the caller.
		 * @param called
		 *            called.
		 * 
		 * @param typesToCall
		 *            the list to check against.
		 */
		TypeChecker(final Type caller, final Type called, final Collection<Type> typesToCall) {
			this.caller = caller;
			this.called = called;
			this.typesToCall = typesToCall;
		}
		
		@Override
		public Boolean handle(final AtomicType atomicType) {
			return ReferencedConstructorsCheck.this.contains(this.caller, atomicType, this.typesToCall);
		}
		
		@Override
		public Boolean handle(final CompositeType compositeType) {
			return compositeType.accept(new CompositeTypeChecker(this.caller, this.called, this.typesToCall));
		}
		
		@Override
		public Boolean handle(final TypeProxy typeProxy) {
			return typeProxy.getTarget().accept(this);
		}
		
	}
	
	/**
	 * Checks if a part of a sum type is part of a given list of types.
	 */
	class CompositeTypeChecker implements CompositeTypeVisitorReturn<Boolean> {
		
		/**
		 * Calling type.
		 */
		private final Type caller;
		/**
		 * Called type.
		 */
		private final Type called;
		/**
		 * types to call.
		 */
		private final Collection<Type> typesToCall;
		
		/**
		 * Returns a new CompositeTypeChecker.
		 * 
		 * @param caller
		 *            the caller.
		 * @param called
		 *            called.
		 * 
		 * @param typesToCall
		 *            the list to check against.
		 */
		CompositeTypeChecker(final Type caller, final Type called, final Collection<Type> typesToCall) {
			this.caller = caller;
			this.called = called;
			this.typesToCall = typesToCall;
		}
		
		@Override
		public Boolean handle(final ListType list) {
			return list.getOf().accept(new TypeChecker(this.caller, this.called, this.typesToCall));
		}
		
		@Override
		public Boolean handle(final MapType map) {
			return map.getOf().accept(new TypeChecker(this.caller, this.called, this.typesToCall));
		}
		
		@Override
		public Boolean handle(final ProductType product) {
			return this.typesToCall.contains(product);
		}
		
		@Override
		public Boolean handle(final SumType sum) {
			final Collection<Type> elements = sum.getElements();
			if (elements.contains(this.caller)) {
				return true;
			}
			final Iterator<Type> parts = sum.getElements().iterator();
			boolean found = false;
			while (!found && parts.hasNext()) {
				found = parts.next().accept(new TypeChecker(this.caller, this.called, this.typesToCall));
			}
			
			final Iterator<Type> superParts = this.typesToCall.iterator();
			while (!found && superParts.hasNext()) {
				found = superParts.next().isSupertypeOf(sum);
			}
			return found;
		}
		
		@Override
		public Boolean handle(final ThrownType thrownType) {
			return this.typesToCall.contains(thrownType);
		}
		
	}
	
	/**
	 * returns true if typesTocall contains atomicType.
	 * 
	 * @param caller
	 *            caller
	 * @param atomicType
	 *            atomicType
	 * @param typesToCall
	 *            typesToCall
	 * @return true if typesTocall contains atomicType. false otherwise.
	 */
	public boolean contains(final Type caller, final AtomicType atomicType, final Collection<Type> typesToCall) {
		final Iterator<Type> types = typesToCall.iterator();
		boolean found = false;
		while (!found && types.hasNext()) {
			found = types.next().accept(new TypeVisitorReturn<Boolean>() {
				
				@Override
				public Boolean handle(final AtomicType atomicType2) {
					return atomicType2.equals(atomicType);
				}
				
				@Override
				public Boolean handle(final CompositeType compositeType) {
					final Collection<Type> typesToCall2 = new LinkedList<>();
					typesToCall2.add(atomicType);
					return compositeType.accept(new CompositeTypeChecker(caller, atomicType, typesToCall2));
				}
				
				@Override
				public Boolean handle(final TypeProxy typeProxy) {
					return typeProxy.getTarget().equals(atomicType);
				}
				
			});
		}
		return found;
	}
	
	/**
	 * Returns the type referenced in state.
	 * 
	 * @param state
	 *            the state to get the type from.
	 * @return the referenced type.
	 * @throws TaskException
	 *             if the state is not a byReference state.
	 */
	protected Type getReferencedType(final ConstructorReferenceState state) throws TaskException {
		return state.accept(new ConstructorReferenceStateVisitorReturnException<Type, TaskException>() {
			
			@Override
			public Type handle(final ConstructorByTypeAndSignatureState byName) throws TaskException {
				throw new TaskException("ByName states of ConstructorReferenceStates are not allowed here!");
			}
			
			@Override
			public Type handle(final ConstructorInvalidState invalid) throws TaskException {
				throw new TaskException("Invalid states of ConstructorReferenceStates are not allowed here!");
			}
			
			@Override
			public Type handle(final ConstructorByReferenceState byReference) throws TaskException {
				return byReference.getType();
			}
			
		});
	}
	
	@Override
	public void finalizeTask() throws TaskException {
		// Nothing to do.
	}
	
	@Override
	public void beginTask() throws TaskException {
		// Nothing to do.
	}
	
}
