package de.fhdw.wtf.walker.tasks;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

import de.fhdw.wtf.common.ast.Attribute;
import de.fhdw.wtf.common.ast.ConstructorOrOperation;
import de.fhdw.wtf.common.ast.Group;
import de.fhdw.wtf.common.ast.Model;
import de.fhdw.wtf.common.ast.type.AtomicType;
import de.fhdw.wtf.common.ast.type.BaseType;
import de.fhdw.wtf.common.ast.type.ClassType;
import de.fhdw.wtf.common.ast.type.Type;
import de.fhdw.wtf.common.ast.visitor.AtomicTypeVisitor;
import de.fhdw.wtf.common.exception.walker.TaskException;
import de.fhdw.wtf.common.task.TaskExecutor;
import de.fhdw.wtf.walker.walker.HelperUtils;
import de.fhdw.wtf.walker.walker.SimpleWalkerTask;

/**
 * This task fill by every Class the List of Subtypes.
 * 
 * This Checker must run after the {@link CyclicInheritanceCheck}.
 */
public final class SubtypesFillTask extends SimpleWalkerTask {
	
	/**
	 * Already found classes with supertypes.
	 */
	private final Collection<ClassType> classesWithSupertypes;
	
	/**
	 * Constructor of {@link SubtypesFillTask}.
	 * 
	 * @param m
	 *            model
	 * @param taskmanager
	 *            taskmanager
	 */
	private SubtypesFillTask(final Model m, final TaskExecutor taskmanager) {
		super(m, taskmanager);
		this.classesWithSupertypes = new ArrayList<>();
	}
	
	/**
	 * Creates a {@link SubtypesFillTask}-Object.
	 * 
	 * @param m
	 *            model
	 * @param taskmanager
	 *            taskmanager
	 * @return The {@link SubtypesFillTask}-Object.
	 */
	public static SubtypesFillTask create(final Model m, final TaskExecutor taskmanager) {
		return new SubtypesFillTask(m, taskmanager);
	}
	
	@Override
	public void handleClass(final ClassType c) throws TaskException {
		if (c.getSuperTypes().size() > 0) {
			this.getClassesWithSupertypes().add(c);
		}
		
		final Iterator<Type> i = c.getSuperTypes().iterator();
		while (i.hasNext()) {
			final Type current = i.next();
			final AtomicType targetType = HelperUtils.getReferencedAtomicType(current);
			targetType.accept(new AtomicTypeVisitor() {
				
				@Override
				public void handle(final ClassType clazz) {
					clazz.getSubTypes().add(c);
				}
				
				@Override
				public void handle(final BaseType baseType) {
					// Nothing
				}
			});
		}
	}
	
	@Override
	public void handleGroup(final Group g) throws TaskException {
		// Nothing to do here.
	}
	
	@Override
	public void handleAttribute(final Attribute a, final ClassType owner) throws TaskException {
		// Nothing to do here.
	}
	
	@Override
	public void finalizeTask() throws TaskException {
		// Nothing to do here.
	}
	
	@Override
	public void beginTask() throws TaskException {
		// Nothing to do here
	}
	
	/**
	 * Returned the List of Classtypes, which have Supertypes.
	 * 
	 * @return List of Classtypes with Supertypes
	 */
	public Collection<ClassType> getClassesWithSupertypes() {
		return this.classesWithSupertypes;
	}
	
	@Override
	public void handleConstructorOrOperation(final ConstructorOrOperation coo, final ClassType owner)
			throws TaskException {
		// Nothing to do here.
		
	}
	
}
