package de.fhdw.wtf.walker.tasks;

import java.util.HashSet;
import java.util.Set;

import de.fhdw.wtf.common.ast.Model;
import de.fhdw.wtf.common.ast.UnqualifiedName;
import de.fhdw.wtf.common.ast.type.AtomicType;
import de.fhdw.wtf.common.ast.type.ByReferenceState;
import de.fhdw.wtf.common.ast.type.CompositeType;
import de.fhdw.wtf.common.ast.type.ListType;
import de.fhdw.wtf.common.ast.type.MapType;
import de.fhdw.wtf.common.ast.type.ProductType;
import de.fhdw.wtf.common.ast.type.SumType;
import de.fhdw.wtf.common.ast.type.ThrownType;
import de.fhdw.wtf.common.ast.type.Type;
import de.fhdw.wtf.common.ast.type.TypeProxy;
import de.fhdw.wtf.common.ast.visitor.CompositeTypeVisitor;
import de.fhdw.wtf.common.ast.visitor.TypeVisitor;
import de.fhdw.wtf.common.constants.referencer.BaseTypeConstants;
import de.fhdw.wtf.common.exception.ast.ASTException;
import de.fhdw.wtf.common.exception.walker.TaskException;
import de.fhdw.wtf.common.task.TaskExecutor;
import de.fhdw.wtf.common.token.DummyToken;
import de.fhdw.wtf.common.token.IdentifierToken;
import de.fhdw.wtf.walker.walker.HelperUtils;
import de.fhdw.wtf.walker.walker.SimpleWalkerTaskForTypes;

/**
 * This task fill <br>
 * - by every PROTOTYPE-Sum the List of Supertypes.<br>
 * - by every PROTOTYPE-Product the abstract Product to the List of Supertypes.<br>
 * - by every Element of a PROTOTYPE-Sum the PROTOTYPE-Sum in the List of Supertypes.
 *
 * This Checker must run after the {@link PrototypesTask}.
 */
public final class SupertypesFillTask extends SimpleWalkerTaskForTypes {
	
	// It could be that some inheritances are duplicated in the list of supertypes
	
	/**
	 * All Sums of the model.
	 */
	private final Set<SumType> foundSums;
	
	/**
	 * All types of the model.
	 */
	private final Set<Type> allTypesOfModel;
	
	/**
	 * A TypeProxy pointing to the Anything type by reference.
	 */
	private TypeProxy typeProxyToAnything;
	
	/**
	 * Constructor of {@link SupertypesFillTask}.
	 *
	 * @param m
	 *            model
	 * @param taskmanager
	 *            taskmanager
	 */
	public SupertypesFillTask(final Model m, final TaskExecutor taskmanager) {
		super(m, taskmanager, true);
		this.foundSums = new HashSet<>();
		this.allTypesOfModel = new HashSet<>();
	}
	
	/**
	 * Creates a {@link SupertypesFillTask}-Object.
	 *
	 * @param m
	 *            model
	 * @param taskmanager
	 *            taskmanager
	 *
	 * @return The {@link SupertypesFillTask}-Object.
	 */
	public static SupertypesFillTask create(final Model m, final TaskExecutor taskmanager) {
		return new SupertypesFillTask(m, taskmanager);
	}
	
	@Override
	public void beginTask() throws TaskException {
		final ByReferenceState byReferenceState =
				ByReferenceState.create(
						this.getModel().getAnything().getPrototype(),
						UnqualifiedName.create(IdentifierToken.create(
								BaseTypeConstants.ANYTHING_CONSTANT,
								DummyToken.getDummyPosition())));
		this.typeProxyToAnything =
				TypeProxy.create(DummyToken.getInstance(), byReferenceState, DummyToken.getInstance());
	}
	
	@Override
	public void handleType(final Type c) throws TaskException {
		c.accept(new TypeVisitor() {
			
			@Override
			public void handle(final TypeProxy typeProxy) {
				// Ignore
			}
			
			@Override
			public void handle(final CompositeType compositeType) {
				
				SupertypesFillTask.this.getAllTypesOfModel().add(compositeType);
				
				compositeType.accept(new CompositeTypeVisitor() {
					
					@Override
					public void handle(final SumType sum) {
						SupertypesFillTask.this.getFoundSums().add(sum);
						// Put the Sum in the list of supertypes of each sum-element
						Type supertype = sum;
						if (sum == SupertypesFillTask.this.getModel().getAnything().getPrototype()) {
							supertype = SupertypesFillTask.this.typeProxyToAnything;
						}
						for (final Type current : sum.getElements()) {
							// current could be TypeProxy!
							HelperUtils.getReferencedType(current).getSuperTypes().add(supertype);
						}
						
					}
					
					@Override
					public void handle(final ProductType product) {
						// Put the abstract Product in the list of supertypes of the product
						if (product != product.getAbstractPrototype()) {
							product.getSuperTypes().add(product.getAbstractPrototype());
						}
					}
					
					@Override
					public void handle(final MapType map) {
						// Nothing
					}
					
					@Override
					public void handle(final ListType list) {
						// Nothing
					}
					
					@Override
					public void handle(final ThrownType thrownType) {
						// Nothing
					}
				});
			}
			
			@Override
			public void handle(final AtomicType atomicType) {
				SupertypesFillTask.this.allTypesOfModel.add(atomicType);
			}
		});
	}
	
	@Override
	public void finalizeTask() throws TaskException {
		// Fill the list of supertypes of the sums
		for (final SumType currentSum : this.getFoundSums()) {
			for (final Type currentOtherType : this.getAllTypesOfModel()) {
				if (currentOtherType.isSupertypeOf(currentSum)) {
					currentOtherType.accept(new TypeVisitor() {
						
						@Override
						public void handle(final TypeProxy typeProxy) {
							throw new ASTException("It couldn't be a TypeProxy");
						}
						
						@Override
						public void handle(final CompositeType compositeType) {
							compositeType.accept(new CompositeTypeVisitor() {
								
								@Override
								public void handle(final SumType sum) {
									if (sum == SupertypesFillTask.this.getModel().getAnything().getPrototype()) {
										currentSum.getSuperTypes().add(SupertypesFillTask.this.typeProxyToAnything);
									} else {
										currentSum.getSuperTypes().add(currentOtherType);
									}
								}
								
								@Override
								public void handle(final ProductType product) {
									currentSum.getSuperTypes().add(currentOtherType);
								}
								
								@Override
								public void handle(final MapType map) {
									currentSum.getSuperTypes().add(currentOtherType);
								}
								
								@Override
								public void handle(final ListType list) {
									currentSum.getSuperTypes().add(currentOtherType);
								}
								
								@Override
								public void handle(final ThrownType thrownType) {
									throw new ASTException("A ThrownType cant be a SuperType of any class.");
								}
							});
						}
						
						@Override
						public void handle(final AtomicType atomicType) {
							final ByReferenceState byReferenceState =
									ByReferenceState.create(currentOtherType, atomicType.getTypeName());
							final TypeProxy typeProxy =
									TypeProxy.create(
											DummyToken.getInstance(),
											byReferenceState,
											DummyToken.getInstance());
							currentSum.getSuperTypes().add(typeProxy);
						}
					});
					
				}
			}
		}
	}
	
	/**
	 * Returns the list of sums, which was found.
	 *
	 * @return sums
	 */
	public Set<SumType> getFoundSums() {
		return this.foundSums;
	}
	
	/**
	 * Returns the list of types, which was found.
	 *
	 * @return types
	 */
	public Set<Type> getAllTypesOfModel() {
		return this.allTypesOfModel;
	}
	
}
